/*
    Copyright (C) 2011 Harald Sitter <sitter@kde.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) version 3, or any
    later version accepted by the membership of KDE e.V. (or its
    successor approved by the membership of KDE e.V.), Nokia Corporation
    (or its successors, if any) and the KDE Free Qt Foundation, which shall
    act as a proxy defined in Section 6 of version 3 of the license.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PHONON_GSTREAMER_VIDEOSINK_H
#define PHONON_GSTREAMER_VIDEOSINK_H

#include <gst/video/gstvideosink.h>

G_BEGIN_DECLS

#define P_GST_TYPE_VIDEO_SINK p_gst_video_sink_get_type()
#define P_GST_VIDEO_SINK(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), P_GST_TYPE_VIDEO_SINK, PGstVideoSink))
#define P_GST_VIDEO_SINK_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), P_GST_TYPE_VIDEO_SINK, PGstVideoSinkClass))
#define P_GST_IS_VIDEO_SINK(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), P_GST_TYPE_VIDEO_SINK))
#define P_GST_IS_VIDEO_SINK_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), P_GST_TYPE_VIDEO_SINK))
#define P_GST_VIDEO_SINK_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), P_GST_TYPE_VIDEO_SINK, PGstVideoSinkClass))

typedef struct _PGstVideoSink PGstVideoSink;
typedef struct _PGstVideoSinkClass PGstVideoSinkClass;

struct _PGstVideoSink {
    GstVideoSink parent;

    void *userData;
    void (*renderCallback)(GstBuffer *, void *);
};

struct _PGstVideoSinkClass {
    GstVideoSinkClass parent_class;
};

GType p_gst_video_sink_get_type(void) G_GNUC_CONST;

GstCaps *p_gst_video_sink_get_static_caps();

G_END_DECLS

#endif /* PHONON_GSTREAMER_VIDEOSINK_H */
